@@ -52,6 +52,8 @@ module Data.Iteratee.Iteratee (
   ,eneeCheckIfDonePass
   ,mergeEnums
   -- ** Enumeratee Combinators
+  ,($=)
+  ,(=$)
   ,(><>)
   ,(<><)
   -- * Misc.
@@ -253,7 +255,7 @@ eneeCheckIfDonePass
      )
   -> Enumeratee elo eli m a
 eneeCheckIfDonePass f = eneeCheckIfDoneHandle (\k e -> f k (Just e)) f
-{-# INLINE eneeCheckIfDonePass #-}
+{-# INLINEABLE eneeCheckIfDonePass #-}
 
 eneeCheckIfDoneIgnore
   :: (Monad m, NullPoint elo)
@@ -420,6 +422,29 @@ enumErr e iter = runIter iter onDone onCont
 (e1 >>> e2) i =  e1 i >>= e2
   -- I think (>>>) is identical to (>=>)...
 
+infixr 0 =$
+
+-- | Combines an Enumeratee from @s@ to @s'@ and an Iteratee that
+--  consumes @s'@ into an Iteratee which consumes @s@
+(=$)
+  :: (Nullable s, Nullable s', Monad m)
+  => Enumeratee s s' m a
+  -> Iteratee s' m a
+  -> Iteratee s m a
+(=$) = (.) joinI
+
+infixl 1 $=
+
+-- | Combines Enumerator which produces stream of @s@ and @Enumeratee@
+--  which transforms stream of @s@ to stream
+--  of @s'@ to into Enumerator which produces stream of @s'@
+($=)
+  :: (Nullable s, Nullable s', Monad m)
+  => (forall a. Enumerator s m a)
+  -> Enumeratee s s' m b
+  -> Enumerator s' m b
+($=) enum enee iter = enum (enee iter) >>= run
+
 
 -- | Enumeratee composition
 -- Run the second enumeratee within the first.  In this example, stream2list
