@@ -30,6 +30,8 @@ module Data.Iteratee.ListLike (
   ,breakE
   ,take
   ,takeUpTo
+  ,takeWhile
+  ,takeWhileE
   ,mapStream
   ,rigidMapStream
   ,filter
@@ -64,7 +66,7 @@ module Data.Iteratee.ListLike (
 )
 where
 
-import Prelude hiding (mapM_, null, head, last, drop, dropWhile, take, break, foldl, foldl1, length, filter, sum, product, zip, zip3, sequence_)
+import Prelude hiding (mapM_, null, head, last, drop, dropWhile, take, takeWhile, break, foldl, foldl1, length, filter, sum, product, zip, zip3, sequence_)
 
 import qualified Prelude as Prelude
 
@@ -318,13 +320,13 @@ breakE
   :: (Monad m, LL.ListLike s el, NullPoint s)
   => (el -> Bool)
   -> Enumeratee s s m a
-breakE cpred = eneeCheckIfDone (liftI . step)
+breakE cpred = eneeCheckIfDonePass (icont . step)
  where
   step k (Chunk s)
       | LL.null s  = liftI (step k)
       | otherwise  = case LL.break cpred s of
         (str', tail')
-          | LL.null tail' -> eneeCheckIfDone (liftI . step) . k $ Chunk str'
+          | LL.null tail' -> eneeCheckIfDonePass (icont . step) . k $ Chunk str'
           | otherwise     -> idone (k $ Chunk str') (Chunk tail')
   step k stream           =  idone (k stream) stream
 {-# INLINE breakE #-}
@@ -412,6 +414,27 @@ takeUpTo i iter
 {-# SPECIALIZE takeUpTo :: Monad m => Int -> Enumeratee B.ByteString B.ByteString m a #-}
 {-# INLINABLE takeUpTo #-}
 
+-- | Takes an element predicate and returns the (possibly empty)
+-- prefix of the stream. All characters
+-- in the string will satisfy the character predicate. If the stream
+-- is not terminated, the first character of the
+-- remaining stream will not satisfy the predicate.
+-- 
+-- The analogue of @List.takeWhile@, see also @break@ and @takeWhileE@
+takeWhile :: (LL.ListLike s el, Monad m) => (el -> Bool) -> Iteratee s m s
+takeWhile = break . (not .)
+{-# INLINEABLE takeWhile #-}
+
+-- |Takes an element predicate and an iteratee, running the iteratee
+-- on all elements of the stream while the predicate is met.
+-- 
+-- This is preferred to @takeWhile@.
+takeWhileE
+ :: (Monad m, LL.ListLike s el, NullPoint s)
+ => (el -> Bool)
+ -> Enumeratee s s m a
+takeWhileE = breakE . (not .)
+{-# INLINEABLE takeWhileE #-}
 
 -- |Map the stream: another iteratee transformer
 -- Given the stream of elements of the type @el@ and the function @(el->el')@,
@@ -481,9 +504,17 @@ group cksz iinit = liftI (step 0 id iinit)
     | otherwise               =
         let (full, rest) = gsplit . mconcat $ pfxd [s]
             pfxd'        = if LL.null rest then id else (rest:)
+            onDone x str = return $ Left (x,str)
+            onCont k Nothing = return . Right . Left . k $ Chunk full
+            onCont k e       = return . Right $ Right (liftI k, e)
         in  do
-              inext <- lift $ enumPure1Chunk full icur
-              liftI $ step (LL.length rest) pfxd' inext
+              res <- lift $ runIter icur onDone onCont
+              case res of
+                Left (x,str) -> idone (idone x str) (Chunk rest)
+                Right (Left inext)  -> liftI $ step (LL.length rest) pfxd' inext
+                Right (Right (inext, e)) -> icont (step (LL.length rest)
+                                                        pfxd' inext)
+                                                  e
   step _ pfxd icur mErr = case pfxd [] of
                          []   -> idone icur mErr
                          rest -> do
@@ -521,8 +552,19 @@ groupBy same iinit = liftI $ go iinit (const True, id)
     -- accumulator.
     go icurr pfx (Chunk s) = case gsplit pfx s of
       ([], partial)   -> liftI $ go icurr partial
-      (full, partial) -> do inext <- lift $ enumPure1Chunk full icurr
-                            liftI $ go inext partial
+      (full, partial) -> do
+        -- if the inner iteratee is done, the outer iteratee needs to be
+        -- notified to terminate.
+        -- if the inner iteratee is in an error state, that error should
+        -- be lifted to the outer iteratee
+        let onCont k Nothing = return $ Right $ Left $ k $ Chunk full
+            onCont k e       = return $ Right $ Right (liftI k, e)
+            onDone x str     = return $ Left (x,str)
+        res <- lift $ runIter icurr onDone onCont
+        case res of
+          Left (x,str) -> idone (idone x str) (Chunk (mconcat $ snd partial []))
+          Right (Left inext)      -> liftI $ go inext partial
+          Right (Right (inext,e)) -> icont (go inext partial) e
     go icurr (_inpfx, pfxd) (EOF mex) = case pfxd [] of
       [] -> lift . enumChunk (EOF mex) $ icurr
       rest -> do inext <- lift . enumPure1Chunk [mconcat rest] $ icurr
@@ -750,7 +792,6 @@ enumPair
   -> Iteratee s m (a, b)
 enumPair = zip
 
-
 -- |Enumerate two iteratees over a single stream simultaneously.
 -- 
 -- Compare to @List.zip@.
@@ -759,25 +800,38 @@ zip
   => Iteratee s m a
   -> Iteratee s m b
   -> Iteratee s m (a, b)
-zip x y = liftI step
+zip x0 y0 = do
+    -- need to check if both iteratees are initially finished.  If so,
+    -- we don't want to push a chunk which will be dropped
+    (a', x') <- lift $ runIter x0 od oc
+    (b', y') <- lift $ runIter y0 od oc
+    case checkDone a' b' of
+      Just (Right (a,b,s))  -> idone (a,b) s  -- 's' may be EOF, needs to stay
+      Just (Left (Left a))  -> liftM (a,) y'
+      Just (Left (Right b)) -> liftM (,b) x'
+      Nothing               -> liftI (step x' y')
   where
-    step (Chunk xs) | nullC xs = liftI step
-    step (Chunk xs) = do
+    step x y (Chunk xs) | nullC xs = liftI (step x y)
+    step x y (Chunk xs) = do
       (a', x') <- lift $ (\i -> runIter i od oc) =<< enumPure1Chunk xs x
       (b', y') <- lift $ (\i -> runIter i od oc) =<< enumPure1Chunk xs y
       case checkDone a' b' of
-        Just (a, b, s) -> idone (a, b) s
-        Nothing        -> zip x' y'
-    step (EOF err) = joinIM $ case err of
+        Just (Right (a,b,s))  -> idone (a,b) s
+        Just (Left (Left a))  -> liftM (a,) y'
+        Just (Left (Right b)) -> liftM (,b) x'
+        Nothing               -> liftI (step x' y')
+    step x y (EOF err) = joinIM $ case err of
       Nothing -> (liftM2.liftM2) (,) (enumEof   x) (enumEof   y)
       Just e  -> (liftM2.liftM2) (,) (enumErr e x) (enumErr e y)
 
     od a s = return (Just (a, s), idone a s)
     oc k e = return (Nothing    , icont k e)
 
-    checkDone r1 r2 =
-      r1 >>= \(a, s1) -> r2 >>= \(b, s2) ->
-      return (a, b, shorter s1 s2)
+    checkDone r1 r2 = case (r1, r2) of
+      (Just (a, s1), Just (b,s2)) -> Just $ Right (a, b, shorter s1 s2)
+      (Just (a, _), Nothing)      -> Just . Left $ Left a
+      (Nothing, Just (b, _))      -> Just . Left $ Right b
+      (Nothing, Nothing)          -> Nothing
 
     shorter c1@(Chunk xs) c2@(Chunk ys)
       | LL.length xs < LL.length ys = c1
@@ -825,7 +879,17 @@ enumWith
   => Iteratee s m a
   -> Iteratee s m b
   -> Iteratee s m (a, b)
-enumWith i1 i2 = go i1 i2
+enumWith i1 i2 = do
+    -- as with zip, first check to see if the initial iteratee is complete,
+    -- otherwise data would be dropped.
+    -- running the second iteratee as well to prevent a monadic effect mismatch
+    -- although I think that would be highly unlikely to happen in common
+    -- code
+    (a', x') <- lift $ runIter i1 od oc
+    (_,  y') <- lift $ runIter i2 od oc
+    case a' of
+      Just (a, s) -> flip idone s =<< lift (liftM (a,) $ run i2)
+      Nothing     -> go x' y'
   where
     od a s = return (Just (a, s), idone a s)
     oc k e = return (Nothing    , icont k e)
